From 91f2c94d333fbd391dbde30eff48ee1e859bbdf8 Mon Sep 17 00:00:00 2001
From: Kevin Darbyshire-Bryant <ldir@darbyshire-bryant.me.uk>
Date: Tue, 5 Mar 2019 14:48:30 +0000
Subject: [PATCH] start of act_connmark hack

Signed-off-by: Kevin Darbyshire-Bryant <ldir@darbyshire-bryant.me.uk>
---
 net/sched/act_connmark.c | 61 ++++++++++++++++++++++++++++++++++++++++
 1 file changed, 61 insertions(+)

diff --git a/net/sched/act_connmark.c b/net/sched/act_connmark.c
index 8475913f2070..033e54dbee28 100644
--- a/net/sched/act_connmark.c
+++ b/net/sched/act_connmark.c
@@ -41,6 +41,7 @@ static int tcf_connmark_act(struct sk_buff *skb, const struct tc_action *a,
 	struct nf_conntrack_zone zone;
 	struct nf_conn *c;
 	int proto;
+	u8 dscp;
 
 	spin_lock(&ca->tcf_lock);
 	tcf_lastuse_update(&ca->tcf_tm);
@@ -62,6 +63,36 @@ static int tcf_connmark_act(struct sk_buff *skb, const struct tc_action *a,
 
 	c = nf_ct_get(skb, &ctinfo);
 	if (c) {
+		if (dscpops & GET_DSCP && proto && !(c->mark & 0x01000000)) {
+			/* mark does not cotnain DSCP so store DSCP bits into c->mark */
+			switch (proto) {
+			case NFPROTO_IPV4:
+				dscp = ipv4_get_dsfield(ip_hdr(skb));
+				break;
+			case NFPROTO_IPV6:
+				dscp = ipv6_get_dsfield(ip_hdr(skb));
+				break;
+			default:
+				dscp = 0;
+				break;
+			}
+			c->mark &= 0x00ffffff;
+			c->mark |= (0x01 | dscp) << 24;
+		} else if (dscpops & SET_DSCP && proto && (c->mark & 0x01000000)) {
+			/* mark contains DSCP so restore DSCP bits from c->mark into diffserv */
+			switch (proto) {
+			case NFPROTO_IPV4:
+				if (ipv4_get_dsfield(ip_hdr(skb) & ~INET_ECN_MASK) != c->mark >> 24 & ~INET_ECN_MASK)
+					ipv4_change_dsfield(ip_hdr(skb), INET_ECN_MASK, c->mark >> 24 & ~INET_ECN_MASK);
+				break;
+			case NFPROTO_IPV6:
+				if (ipv6_get_dsfield(ip_hdr(skb) & ~INET_ECN_MASK) != c->mark >> 24 & ~INET_ECN_MASK)
+					ipv6_change_dsfield(ip_hdr(skb), INET_ECN_MASK, c->mark >> 24 & ~INET_ECN_MASK);
+				break;
+			default:
+				break;
+			}
+		}
 		skb->mark = c->mark;
 		/* using overlimits stats to count how many packets marked */
 		ca->tcf_qstats.overlimits++;
@@ -82,6 +113,36 @@ static int tcf_connmark_act(struct sk_buff *skb, const struct tc_action *a,
 	c = nf_ct_tuplehash_to_ctrack(thash);
 	/* using overlimits stats to count how many packets marked */
 	ca->tcf_qstats.overlimits++;
+	if (dscpops & GET_DSCP && proto && !(c->mark & 0x01000000)) {
+		/* store the DSCP bits into c->mark */
+		switch (proto) {
+		case NFPROTO_IPV4:
+			dscp = ipv4_get_dsfield(ip_hdr(skb));
+			break;
+		case NFPROTO_IPV6:
+			dscp = ipv6_get_dsfield(ip_hdr(skb));
+			break;
+		default:
+			dscp = 0;
+			break;
+		}
+		c->mark &= 0x00ffffff;
+		c->mark |= (0x01 | dscp) << 24;
+	} else if (dscpops & SET_DSCP && proto && (c->mark & 0x01000000)) {
+		/* restore the DSCP bits from c->mark into diffserv */
+		switch (proto) {
+		case NFPROTO_IPV4:
+			if (ipv4_get_dsfield(ip_hdr(skb) & ~INET_ECN_MASK) != c->mark >> 24 & ~INET_ECN_MASK)
+				ipv4_change_dsfield(ip_hdr(skb), INET_ECN_MASK, c->mark >> 24 & ~INET_ECN_MASK);
+			break;
+		case NFPROTO_IPV6:
+			if (ipv6_get_dsfield(ip_hdr(skb) & ~INET_ECN_MASK) != c->mark >> 24 & ~INET_ECN_MASK)
+				ipv6_change_dsfield(ip_hdr(skb), INET_ECN_MASK, c->mark >> 24 & ~INET_ECN_MASK);
+			break;
+		default:
+			break;
+		}
+	}
 	skb->mark = c->mark;
 	nf_ct_put(c);
 
-- 
2.17.2 (Apple Git-113)

